"""
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
"""
qute is an extension to the Qt.py module written by Marcus Ottosson and
can be found here:
https://github.com/mottosso/Qt.py

This extension aims to expose additional help utilities to Qt which are
common place within coding projects.
"""
__author__ = "Michael Malinowski"
__copyright__ = "Copyright (C) 2019 Michael Malinowski"
__license__ = "MIT"
__version__ = "1.0.10"

# -- Import all our Qt variables into this namespace - which
# -- makes it trivial to use later
from .vendor.Qt.QtCore import *
from .vendor.Qt.QtGui import *
from .vendor.Qt.QtWidgets import *
#
# from .vendor.Qt import QtWidgets
# print(sorted(dir(QtWidgets)))
# -- We import QtCompat explicitly
from .vendor.Qt import QtCompat

from ._utils import slimify
from ._utils import qApp
from ._utils import toGrayscale
from ._utils import emptyLayout
from ._utils import addLabel

from ._style import applyStyle
from ._style import getCompoundedStylesheet

from ._menu import menuFromDictionary

from ._derive import deriveWidget
from ._derive import deriveValue
from ._derive import setBlindValue
from ._derive import connectBlind

from ._windows import mainWindow

from ._ui import loadUi

from . import constants





"""
This module attempts to access widgets and values in a blind manor - allowing
you to extract the most common forms of data from them. This is particularly
useful when you want to represent values through an interface without hard-
coding what they are.
"""
from .vendor.Qt import QtWidgets
import types


_NUMERIC_UI_MAX = 9*9*9
_NUMERIC_UI_MIN = _NUMERIC_UI_MAX * -1


# ------------------------------------------------------------------------------
def is_string(value):
    """
    Py2/3 compatible string testing

    :param value:
    :return:
    """
    try:
        return isinstance(value, basestring)
    except NameError:
        return isinstance(value, str)


# ------------------------------------------------------------------------------
def deriveWidget(value, label=''):
    """
    Given the data type of the value, this will make a guess at the best
    fit ui element to represent that value.
    
    :param value: The value you want to represent through a widget
    :type value: bool, int, float, str, list or dict.

    :param label: This is how any labelled widget will be displayed as
    :type label: str

    :return: QWidget
    """
    # -- Determine the type of the value coming in
    value_type = type(value)

    if value_type is bool:
        derived = QtWidgets.QCheckBox(label)
        derived.setChecked(value)
        return derived

    if is_string(value):
        derived = QtWidgets.QLineEdit()
        derived.setText(value)
        return derived

    if value_type is float:
        derived = QtWidgets.QDoubleSpinBox()
        derived.setMaximum(_NUMERIC_UI_MAX)
        derived.setMinimum(_NUMERIC_UI_MIN)
        derived.setValue(value)
        return derived

    if value_type is int:
        derived = QtWidgets.QSpinBox()
        derived.setMaximum(_NUMERIC_UI_MAX)
        derived.setMinimum(_NUMERIC_UI_MIN)
        derived.setValue(value)
        return derived

    if value_type in [list, tuple]:
        derived = QtWidgets.QComboBox()

        for item in value:
            derived.addItem(item)

        return derived

    elif value_type is dict:
        widget = QtWidgets.QComboBox()

        for k, v in value.items():
            widget.addItem(k, userData=v)
        return widget

    else:
        return None


# ------------------------------------------------------------------------------
def deriveValue(widget):
    """
    Given a QWidget it will call the widgets specific method to return 
    the likely value represented by that widget.
    
    :param widget: QWidget to derive the value from
    :type widget: QWidget

    :return: dependent on the widget
    """
    if isinstance(widget, QtWidgets.QAbstractSpinBox):
        return widget.value()

    if isinstance(widget, QtWidgets.QLineEdit):
        return widget.text()

    if isinstance(widget, QtWidgets.QComboBox):
        data = widget.itemData(widget.currentIndex())
        return data or widget.currentText()

    if isinstance(widget, QtWidgets.QAbstractButton):
        return widget.isChecked() if widget.isCheckable() else widget.isDown()

    return None


# ------------------------------------------------------------------------------
def setBlindValue(widget, value):
    """
    This is a blind setter method, where the type of widget is unknown and
    will therefore be handled within the function.

    :param widget: The widget you want to set the value for
    :type widget: QWidget

    :param value: The value you want to apply to the widget
    :type value: variable

    :return: True if a value was set 
    """
    if isinstance(widget, QtWidgets.QAbstractSpinBox):
        widget.setValue(float(value))
        return True

    if isinstance(widget, QtWidgets.QLineEdit):
        widget.setText(str(value))
        return True

    if isinstance(widget, QtWidgets.QComboBox):
        for i in range(widget.count()):
            if widget.itemText(i) == str(value):
                widget.setCurrentIndex(i)
                return True

    if isinstance(widget, QtWidgets.QAbstractButton):
        widget.setChecked(bool(value))
        return True

    return False


# ------------------------------------------------------------------------------
def connectBlind(widget, callback):
    """
    This is a blind approach to connecting the most likely value change
    event of a given widget to the given callback.
    
    Note: Because of the nature of signals, they can give a variety of 
    different arguments during the signal call. For that reason it is 
    highly recommended to utilise *args, **kwargs within the callback 
    if you do not know exactly which signal will be connected.
    
    :param widget: The widget expected to emit the change signal
    :type widget: QWidget

    :param callback: The callback/function which should be called upon
        a value changed event
    :type callback: callable object

    :return: True on successful connection, False otherwise.
    """
    if isinstance(widget, QtWidgets.QLineEdit):
        widget.textChanged.connect(callback)
        return True

    if isinstance(widget, QtWidgets.QAbstractSpinBox):
        widget.valueChanged.connect(callback)
        return True

    if isinstance(widget, QtWidgets.QComboBox):
        widget.currentIndexChanged.connect(callback)
        return True

    if isinstance(widget, QtWidgets.QAbstractButton):
        widget.clicked.connect(callback)
        return True

    return False



"""
This holds a set of small helper functions
"""
from .vendor import Qt


# ------------------------------------------------------------------------------
def addLabel(widget, label):
    """
    Adds a label to the widget, returning a layout with the label 
    on the left and the widget on the right.

    :param widget: Widget to be given a label
    :type widget: QWidget

    :param label: Label text to show
    :type label: str

    :return: SlimHBoxLayout
    """
    label = Qt.QtWidgets.QLabel(label)
    layout = slimify(Qt.QtWidgets.QHBoxLayout())

    layout.addWidget(label)
    layout.addSpacerItem(
        Qt.QtWidgets.QSpacerItem(
            10,
            0,
            Qt.QtWidgets.QSizePolicy.Expanding,
            Qt.QtWidgets.QSizePolicy.Fixed,
        ),
    )
    layout.addWidget(widget)

    layout.setStretch(
        1,
        1,
    )

    return layout


# --------------------------------------------------------------------------
def emptyLayout(layout):
    """
    Clears the layout of all its children, removing them entirely.

    :param layout: The layout to empty.
    :type layout: QLayout

    :param recurse: If True, clear contents recursively.
        Default True.
    :type recurse: bool

    :return: None
    """
    for i in reversed(range(layout.count())):
        item = layout.takeAt(i)

        if isinstance(item, Qt.QtWidgets.QWidgetItem):
            widget = item.widget()
            widget.setParent(None)
            widget.deleteLater()

        elif isinstance(item, Qt.QtWidgets.QSpacerItem):
            pass

        else:
            emptyLayout(item.layout())


# ------------------------------------------------------------------------------
def toList(given):
    """
    This will take what is given and wrap it in a list if it is not already
    a list, otherwise it will simply return what it has been given.
     
    :return: list()
    """
    if not isinstance(given, (tuple, list)):
        given = [given]

    return given


# ------------------------------------------------------------------------------
def slimify(layout):
    # -- Apply the formatting
    layout.setContentsMargins(
        *[0, 0, 0, 0]
    )
    layout.setSpacing(0)

    return layout


# --------------------------------------------------------------------------
def toGrayscale(pixmap):
    """
    Creates a new pixmap which is a grayscale version of the given
    pixmap

    :param pixmap: QPixmap

    :return: QPixmap
    """
    # -- Get an image object
    image = pixmap.toImage()

    # -- Cycle the pixels and convert them to grayscale
    for x in range(image.width()):
        for y in range(image.height()):

            # -- Grayscale the pixel
            gray = Qt.QtGui.qGray(
                image.pixel(
                    x,
                    y,
                ),
            )

            # -- Set the pixel back into the image
            image.setPixel(
                x,
                y,
                Qt.QtGui.QColor(
                    gray,
                    gray,
                    gray,
                ).rgb()
            )

    # -- Re-apply the alpha channel
    image.setAlphaChannel(
        pixmap.toImage().alphaChannel()
    )

    # -- Return the pixmap
    return Qt.QtGui.QPixmap.fromImage(image)


# ------------------------------------------------------------------------------
def qApp(*args, **kwargs):
    """
    This will return the QApplication instance if one is available, otherwise
    it will create one
    
    :return: QApplication Instance 
    """
    q_app = Qt.QtWidgets.QApplication.instance()

    if not q_app:
        q_app = Qt.QtWidgets.QApplication([])

    return q_app

